/*
	php_zint - ZINT Barcode Generator php extension

	Copyright (c) 2008 Jeong Young Hun()

	This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

	Contact: ka7713@naver.com
*/

#define _USE_32BIT_TIME_T 1
#define HAVE_STDLIB_H 1

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php_zint.h"
#include "php_open_temporary_file.h"

#ifdef _DEBUG
#  define CRTDBG_MAP_ALLOC
#  include <crtdbg.h>
#endif

static unsigned int barcodeType[] = {
	1
	,2
	,3
	,4
	,6
	,7
	,8
	,9
	,13
	,16
	,18
	,20
	,21
	,22
	,23
	,25
	,28
	,29
	,30
	,31
	,32
	,34
	,37
	,40
	,47
	,49
	,50
	,51
	,52
	,53
	,55
	,56
	,57
	,58
	,60
	,63
	,66
	,67
	,68
	,69
	,70
	,71
	,72
	,75
	,79
	,80
	,81
	,82
	,84
	,85
	,86
	,100
	,101
	,102
	,103
	,104
	,105
	,106
	,110
	,111
	,112
	,113
	,114
	,115
	,116
	,117
	,118
	,119
	,120
};

static int linearSearch( const int array[], int key, int size );
static char *getTempfile();

PHP_FUNCTION(zint_imagepng);

zend_function_entry zint_functions[] = {
	PHP_FE(zint_imagepng, NULL)
	/* __function_entries_here__ */
	{NULL, NULL, NULL}	/* Must be the last line in dmtx_functions[] */
};

zend_module_entry zint_module_entry = {
	STANDARD_MODULE_HEADER,
	PHP_ZINT_EXTNAME,
	zint_functions,
	PHP_MINIT(zint),
	PHP_MSHUTDOWN(zint),
	PHP_RINIT(zint),		/* Replace with NULL if there's nothing to do at request start */
	PHP_RSHUTDOWN(zint),	/* Replace with NULL if there's nothing to do at request end */
	PHP_MINFO(zint),
	PHP_ZINT_VERSION, /* Replace with version number for your extension */
	STANDARD_MODULE_PROPERTIES
};

ZEND_GET_MODULE(zint)

PHP_MINIT_FUNCTION(zint)
{
	REGISTER_LONG_CONSTANT( "ZINT_BARCODE_BIND", 2, CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT( "ZINT_BARCODE_BOX", 4, CONST_CS | CONST_PERSISTENT);

	return SUCCESS;
}

PHP_MSHUTDOWN_FUNCTION(zint)
{
	return SUCCESS;
}

PHP_RINIT_FUNCTION(zint)
{
	return SUCCESS;
}

PHP_RSHUTDOWN_FUNCTION(zint)
{
	return SUCCESS;
}

PHP_MINFO_FUNCTION(zint)
{
	php_info_print_table_start();
	php_info_print_table_row(2, "zint support", "enabled");
	php_info_print_table_row(2, "version", PHP_ZINT_VERSION );
	php_info_print_table_row(2, "libzint version", PHP_ZINT_LIB_VERSION );
	php_info_print_table_row(2, "libpng version", "1.2.32");
	php_info_print_table_end();
}

/*
//options
$params = array(
   'outputfile'		=> string (default NULL,   breturn )
   'barcode'		=> int    (default 1, Number of barcode type)
   'breturn'		=> bool   (default false, outputfile Ǿ  õ)
   'data'			=> string (Barcode content) require
   'height'			=> int    (default 0(auto), 1~1000 pixels, Height of barcode)
   'Whitespace'		=> int    (default 0, 0~1000 pixels, Width of whitespace)
   'border'			=> int    (default 0, 1~1000 pixels, Width of border)
   'outputoption'	=> int [ BARCODE_BIND | BARCODE_BOX ]   (default 0, BARCODE_BIND=border ŭ Ʒ , BARCODE_BOX=border ŭ ڽ)
   'reverse'		=> bool (default false, Reverse colours )
   'fg'				=> string (default 000000, 6 length string, Specify a foreground color)
   'bg'				=> string (default ffffff, 6 length string, Specify a background color)
   'rotate'			=> int (default 0, 90 180 270)
   'cols'			=> int (default 0, 1~30, PDF417 only)
   'vers'			=> int (default 0, 1~9, QR Code only )
   'secure'			=> int (default -1, 1~8, PDF417 and QR Code Error correction level)
   'primary'		=> string (default NULL, Maxicode and Composite Structured primary message)
   'mode'			=> int (1~6, Maxicode and Composite Set encoding mode)
  );
*/

PHP_FUNCTION(zint_imagepng)
{
	char *barcodedata=NULL;
	struct zint_symbol *my_symbol=NULL;
	HashPosition pos;
	zval *arr=NULL, **entry=NULL;
	char *string_key;
	int str_key_len, errornum, rotate_angle=0;
	long option;
	BOOL bSetfile=FALSE, breturn=FALSE, bReverse=FALSE;

#ifdef _DEBUG
	_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
	//_CrtSetBreakAlloc(49);
#endif

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "a", &arr) == FAILURE) {
		return;
	}

	my_symbol = ZBarcode_Create();

	zend_hash_internal_pointer_reset_ex(Z_ARRVAL_P(arr), &pos);
	while (zend_hash_get_current_data_ex(Z_ARRVAL_P(arr), (void **)&entry, &pos) == SUCCESS) {
		zend_hash_get_current_key_ex(Z_ARRVAL_P(arr), &string_key, &str_key_len, &option, 0, &pos);
		
		//BARCODE
		if (0 == stricmp(string_key,"BARCODE")) {
			if (Z_TYPE_PP(entry) == IS_LONG || is_numeric_string(Z_STRVAL_PP(entry), Z_STRLEN_PP(entry), NULL, NULL, 0) == IS_LONG) {
				convert_to_long_ex(entry);
				if (-1 != linearSearch(barcodeType, Z_LVAL_PP(entry), sizeof(barcodeType)/sizeof(unsigned int))) {
					my_symbol->symbology = Z_LVAL_PP(entry);
				}else {
					php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid barcode type");
					goto free;
				}
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid barcode paramete (only integer)");
				goto free;
			}
		//DATA
		}else if (0 == stricmp(string_key,"DATA")) {
			convert_to_string_ex(entry);
			if (0<Z_STRLEN_PP(entry)) {
				barcodedata = Z_STRVAL_PP(entry);
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "data is empty");
				goto free;
			}
		//OUTPUTFILE
		}else if (0 == stricmp(string_key,"OUTPUTFILE")) {
			convert_to_string_ex(entry);
			if (256<=Z_STRLEN_PP(entry)) {
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "outputfile path string too long (max 255)");
				goto free;
			}
			if (0<Z_STRLEN_PP(entry)) {
				strncpy(my_symbol->outfile, Z_STRVAL_PP(entry), Z_STRLEN_PP(entry));
				bSetfile = TRUE;
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "outputfile path is emtpy");
				goto free;
			}
		//BRETURN
		}else if (0 == stricmp(string_key,"BRETURN")) {
			convert_to_boolean_ex(entry);
			if (Z_BVAL_PP(entry)){
				breturn = TRUE;
			}else {
				breturn = FALSE;
			}
		//HEIGHT
		}else if (0 == stricmp(string_key,"HEIGHT")) {
			if (Z_TYPE_PP(entry) == IS_LONG || is_numeric_string(Z_STRVAL_PP(entry), Z_STRLEN_PP(entry), NULL, NULL, 0) == IS_LONG) {
				convert_to_long_ex(entry);
				if((Z_LVAL_PP(entry) >= 1) && (Z_LVAL_PP(entry) <= 1000)) {
					my_symbol->height = Z_LVAL_PP(entry);
				} else {
					php_error_docref(NULL TSRMLS_CC, E_WARNING, "Symbol height out of range" );
					goto free;
				}
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid height paramete (only integer)");
				goto free;
			}
		//Whitespace
		}else if (0 == stricmp(string_key,"WHITESPACE")) {
			if (Z_TYPE_PP(entry) == IS_LONG || is_numeric_string(Z_STRVAL_PP(entry), Z_STRLEN_PP(entry), NULL, NULL, 0) == IS_LONG) {
				convert_to_long_ex(entry);
				if((Z_LVAL_PP(entry) >= 0) && (Z_LVAL_PP(entry) <= 1000)) {
					my_symbol->whitespace_width = Z_LVAL_PP(entry);
				} else {
					php_error_docref(NULL TSRMLS_CC, E_WARNING, "whitespace value out of range (0~1000)" );
					goto free;
				}
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid whitespace paramete (only integer)");
				goto free;
			}
		//border
		}else if (0 == stricmp(string_key,"BORDER")) {
			if (Z_TYPE_PP(entry) == IS_LONG || is_numeric_string(Z_STRVAL_PP(entry), Z_STRLEN_PP(entry), NULL, NULL, 0) == IS_LONG) {
				convert_to_long_ex(entry);
				if((Z_LVAL_PP(entry) >= 0) && (Z_LVAL_PP(entry) <= 1000)) {
					my_symbol->border_width = Z_LVAL_PP(entry);
				} else {
					php_error_docref(NULL TSRMLS_CC, E_WARNING, "Border width out of range (0~1000)" );
					goto free;
				}
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid border paramete (only integer)");
				goto free;
			}
		//outputoption
		}else if (0 == stricmp(string_key,"OUTPUTOPTION")) {
			convert_to_long_ex(entry);
			if (BARCODE_BOX == Z_LVAL_PP(entry) || BARCODE_BIND == Z_LVAL_PP(entry)){
				my_symbol->output_options = Z_LVAL_PP(entry);
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid outputoption (only integer)" );
				goto free;
			}
		//reverse
		}else if (0 == stricmp(string_key,"REVERSE")) {
			convert_to_boolean_ex(entry);
			if (Z_BVAL_PP(entry)){
				bReverse = TRUE;
			}else {
				bReverse = FALSE;
			}
		//fg
		}else if (0 == stricmp(string_key,"FG")) {
			convert_to_string_ex(entry);
			if (6==Z_STRLEN_PP(entry)) {
				strncpy(my_symbol->fgcolour, Z_STRVAL_PP(entry), 7);
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid fg paramete");
				goto free;
			}
		//bg
		}else if (0 == stricmp(string_key,"BG")) {
			convert_to_string_ex(entry);
			if (6==Z_STRLEN_PP(entry)) {
				strncpy(my_symbol->bgcolour, Z_STRVAL_PP(entry), 7);
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid bg paramete");
				goto free;
			}
		//rotate
		}else if (0 == stricmp(string_key,"ROTATE")) {
			convert_to_long_ex(entry);
			//0 90 180 270
			//if (0 <= Z_LVAL_PP(entry) && 360 > Z_LVAL_PP(entry)){
			if (90 == Z_LVAL_PP(entry) || 180 == Z_LVAL_PP(entry) || 270 == Z_LVAL_PP(entry)){
				rotate_angle = Z_LVAL_PP(entry);
			}else{
				rotate_angle = 0;
			}
		//cols
		}else if (0 == stricmp(string_key,"COLS")) {
			if (Z_TYPE_PP(entry) == IS_LONG || is_numeric_string(Z_STRVAL_PP(entry), Z_STRLEN_PP(entry), NULL, NULL, 0) == IS_LONG) {
				convert_to_long_ex(entry);
				if((Z_LVAL_PP(entry) >= 1) && (Z_LVAL_PP(entry) <= 30)) {
					my_symbol->option_2 = Z_LVAL_PP(entry);
				} else {
					php_error_docref(NULL TSRMLS_CC, E_WARNING, "the cols is out of range 1~30" );
					goto free;
				}
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid cols paramete");
				goto free;
			}
		//vers
		}else if (0 == stricmp(string_key,"VERS")) {
			if (Z_TYPE_PP(entry) == IS_LONG || is_numeric_string(Z_STRVAL_PP(entry), Z_STRLEN_PP(entry), NULL, NULL, 0) == IS_LONG) {
				convert_to_long_ex(entry);
				if((Z_LVAL_PP(entry) >= 1) && (Z_LVAL_PP(entry) <= 9)) {
					my_symbol->option_2 = Z_LVAL_PP(entry);
				} else {
					php_error_docref(NULL TSRMLS_CC, E_WARNING, "vers(QR Code version) is out of range 1~9" );
					goto free;
				}
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid vers(QR Code version) (only interger)");
				goto free;
			}
		//secure
		}else if (0 == stricmp(string_key,"SECURE")) {
			if (Z_TYPE_PP(entry) == IS_LONG || is_numeric_string(Z_STRVAL_PP(entry), Z_STRLEN_PP(entry), NULL, NULL, 0) == IS_LONG) {
				convert_to_long_ex(entry);
				if((Z_LVAL_PP(entry) >= 1) && (Z_LVAL_PP(entry) <= 8)) {
					my_symbol->option_1 = Z_LVAL_PP(entry);
				} else {
					php_error_docref(NULL TSRMLS_CC, E_WARNING, "secure(ECC level) is out of range 1~8" );
					goto free;
				}
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid secure(ECC level)");
				goto free;
			}
		//primary
		}else if (0 == stricmp(string_key,"PRIMARY")) {
			convert_to_string_ex(entry);
			if (90 < Z_STRLEN_PP(entry)){
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Primary data string too long (max 90)" );
				goto free;
			}
			if (0 < Z_STRLEN_PP(entry)) {
				strcpy(my_symbol->primary,  Z_STRVAL_PP(entry) );
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Parameter primary: non-empty value required" );
				goto free;
			}
		//mode
		}else if (0 == stricmp(string_key,"MODE")) {
			if (Z_TYPE_PP(entry) == IS_LONG || is_numeric_string(Z_STRVAL_PP(entry), Z_STRLEN_PP(entry), NULL, NULL, 0) == IS_LONG) {
				convert_to_long_ex(entry);
				if((Z_LVAL_PP(entry) >= 1) && (Z_LVAL_PP(entry) <= 6)) {
					my_symbol->option_1 = Z_LVAL_PP(entry);
				} else {
					php_error_docref(NULL TSRMLS_CC, E_WARNING, "mode is out of range 1~6" );
					goto free;
				}
			}else{
				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid mode (only integer)");
				goto free;
			}
		}

		//next array
		zend_hash_move_forward_ex(Z_ARRVAL_P(arr), &pos);
	}

	//Ϸ  ƴϸ ӽ 
	if (!bSetfile) {
		char *temp_outputfile = getTempfile();
		strncpy( my_symbol->outfile, temp_outputfile, strlen(temp_outputfile) );
		strcat( my_symbol->outfile, ".png" );
		efree(temp_outputfile);
		temp_outputfile = NULL;
	}

	//reverse
	if (bReverse) {
		char color_temp[10];

		strcpy(color_temp, my_symbol->fgcolour);
		strcpy(my_symbol->fgcolour, my_symbol->bgcolour);
		strcpy(my_symbol->bgcolour, color_temp);
	}

	//printf( " BARCODE = %d \n DATA = %s \n OUTPUTFILE = %s \n BRETURN = %d\n HEIGHT = %d\n WHITESPACE = %d\n BORDER = %d\n OUTPUTOPTION = %d\n REVERSE = %d\n FG = %s\n BG = %s\n ROTATE = %d\n COLS | VERS = %d\n SECURE | MODE = %d\n PRIMARY = %s\n --\n"
	//	, my_symbol->symbology
	//	, barcodedata
	//	, my_symbol->outfile
	//	, breturn
	//	, my_symbol->height
	//	, my_symbol->whitespace_width 
	//	, my_symbol->border_width 
	//	, my_symbol->output_options
	//	, bReverse
	//	, my_symbol->fgcolour
	//	, my_symbol->bgcolour
	//	, rotate_angle
	//	, my_symbol->option_2
	//	, my_symbol->option_1
	//	, my_symbol->primary
	//);

	//ڵ 
	if (0 == rotate_angle) {
		ZBarcode_Encode_and_Print(my_symbol, barcodedata );
	}else{
		ZBarcode_Encode_and_Print_Rotated(my_symbol, barcodedata, rotate_angle);
	}

	if(strcmp(my_symbol->errtxt, "")) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, my_symbol->errtxt );
		goto free;
	}

	//str ȯ
	if (!bSetfile && breturn) {
		char  *buf=NULL;
		FILE *fp=NULL;
		struct stat st;
		
		fp = fopen(my_symbol->outfile, "rb");

		if (VCWD_STAT(my_symbol->outfile, &st) == -1) {
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Bad file file descriptor");
		}

		buf = safe_emalloc(1, st.st_size, 0);
		fread(buf, st.st_size, 1, fp);
		buf[st.st_size] = '\0';

		RETVAL_STRINGL(buf, st.st_size, 1);

		fclose(fp);
		efree(buf);
		buf = NULL;
	}
	//ٷ 
	else if(!bSetfile) {
		char  buf[4096];
		FILE *fp=NULL;
		int rsize;

		fp = fopen(my_symbol->outfile, "rb");

		while ((rsize = (uint)fread(buf, 1, sizeof(buf), fp)) > 0) {
			php_write(buf, rsize TSRMLS_CC);
		}

		fclose(fp);
		RETVAL_NULL();
	}else {
		RETVAL_NULL();
	}

free:

	if (my_symbol) {
		//ӽ Ǵ   
		//  ƴ 쿡
		if (!bSetfile && 0 == VCWD_ACCESS( my_symbol->outfile, F_OK)){
			VCWD_UNLINK(my_symbol->outfile);
		}
		ZBarcode_Delete(my_symbol);
		my_symbol=NULL;
	}

}

static char *getTempfile()
{
	char *opened_path;
	int fd;

	TSRMLS_FETCH();

	if (-1==(fd = php_open_temporary_fd(NULL, NULL, &opened_path TSRMLS_CC))) {
		return NULL;
	} else {
		close(fd);
		return opened_path;
		
	}
}

static int linearSearch( const int array[], int key, int size )
{
   int n;
   for ( n = 0; n < size; ++n ) {
      if ( array[ n ] == key ) { 
         return n;
      } 
   } 
   return -1;
}